from .resource_balancer import BoundedResourceBalancer
from threading import Thread
from unittest import TestCase


class AsyncRunner(Thread):
    def __init__(self, target):
        def main():
            try:
                self.result = target()
            except BaseException as e:
                self.exception = e

        super().__init__(target=main)
        super().start()

    def join(self, timeout=None):
        super().join(timeout)
        if self.is_alive():
            raise TimeoutError()
        if hasattr(self, "exception"):
            raise self.exception
        else:
            return self.result


class ResourceBalancerTest(TestCase):
    def test_basic_function(self):
        b = BoundedResourceBalancer()
        b.provision([1, 2])

        with b.acquire() as resource1:
            with b.acquire() as resource2:
                assert resource1 != resource2
            with b.acquire() as resource2:
                assert resource1 != resource2
                with b.acquire() as resource3:
                    assert resource3 in (resource1, resource2)
                    with b.acquire() as resource4:
                        assert resource3 != resource4

    def test_delayed_init(self):
        b = BoundedResourceBalancer()

        t = AsyncRunner(b.acquire)
        with self.assertRaises(TimeoutError):
            t.join(0.001)

        b.provision([1])
        assert t.join().__enter__() == 1

    def test_ghosts(self):
        b = BoundedResourceBalancer()
        b.provision([1])

        resource1 = b.acquire()
        with resource1:
            b.provision([])

            t = AsyncRunner(b.acquire)
            with self.assertRaises(TimeoutError):
                t.join(0.001)

            b.provision([1])
            resource2 = t.join()

            assert resource1._instance is resource2._instance
            assert resource2._instance.usage == 2
            assert resource2._instance.resource == 1

    def test_vanishing_resource(self):
        b = BoundedResourceBalancer()
        b.provision([1])

        with b.acquire():
            b.provision([])
            assert b._ghosts[0].usage == 1
            assert b._ghosts[0].resource == 1

        assert b._ghosts[0].usage == 0
        b.provision([])
        assert b._ghosts == []

    def test_even_load(self):
        b = BoundedResourceBalancer()
        b.provision([1, 2])

        with b.acquire() as resource1:
            pass

        with b.acquire() as resource2:
            pass

        assert resource1 != resource2



#!/usr/bin/env python3

from http import HTTPStatus
from logging import getLogger
from signal import SIGTERM, signal
from socket import (
    AF_INET,
    IPPROTO_TCP,
    SHUT_RDWR,
    SO_KEEPALIVE,
    SOCK_STREAM,
    SOL_SOCKET,
    TCP_KEEPCNT,
    TCP_KEEPIDLE,
    TCP_KEEPINTVL,
    getaddrinfo,
    socket,
)
from threading import Thread, Timer
from time import sleep
from wsgiref import simple_server

from werkzeug.serving import WSGIRequestHandler

from .gating import wait_for_idle_cpus
from .resource_balancer import BoundedResourceBalancer, ResourceLoan

LOGGER = getLogger("pothead.worker")


class Handler(WSGIRequestHandler):
    protocol_version = "HTTP/1.1"

    def __init__(self, socket, address, server, cb_request_received):
        self.cb_request_received = cb_request_received
        self.env = None
        super().__init__(socket, address, server)

    def handle_one_request(self):
        try:
            # Handler may be used for several requests, so we must reset
            # environment overrides each time
            self.env = None
            super().handle_one_request()
        finally:
            self.close_connection = True

    def make_environ(self):
        environ = super().make_environ()

        if self.cb_request_received:
            self.cb_request_received(environ)

        return environ

    def parse_request(self):
        success = super().parse_request()

        # This socket shall no longer be fast-closed on SIGTERM
        socket = self.request
        self.server.sockets_waiting_for_request.remove(socket)

        return success

    def log_message(self, format, *args):
        LOGGER.info(format, *args)

    def log_request(self, code="-", size="-"):
        if isinstance(code, HTTPStatus):
            code = code.value
        LOGGER.info('"%s" %s %s', self.requestline, str(code), str(size))

    def log_error(self, format, *args):
        LOGGER.error(format, *args)


class LoadBalancer:
    REFRESH_INTERVAL = 5

    def __init__(self, host, port):
        self.host = host
        self.port = port
        self._balancer = BoundedResourceBalancer(1)
        self.refresh()

    def acquire_addr(self) -> ResourceLoan:
        return self._balancer.acquire()

    def refresh(self):
        try:
            results = getaddrinfo(self.host, self.port, type=SOCK_STREAM)
            hosts = (sockaddr for _f, _t, _p, _c, sockaddr in results)
            self._balancer.provision(hosts)
        except BaseException:
            LOGGER.exception(
                "Failed to refresh endpoints for %s:%d", self.host, self.port
            )
        t = Timer(self.REFRESH_INTERVAL, self.refresh)
        t.setDaemon(True)
        t.start()


class Server:
    ssl_context = None
    multithread = False
    multiprocess = False
    server_address = "localhost"
    passthrough_errors = False
    shutdown_signal = False
    running = True

    def __init__(self, addr, app, load_balancer=LoadBalancer):
        (host, port) = addr
        # Set up base environment
        env = self.base_environ = {}
        env["SERVER_NAME"] = host
        env["GATEWAY_INTERFACE"] = "HTTP/1.1"
        env["SERVER_PORT"] = port
        env["REMOTE_HOST"] = ""
        env["CONTENT_LENGTH"] = ""
        env["SCRIPT_NAME"] = ""

        self.broker_balancer = load_balancer(host, port)
        self.app = app
        self.sockets_waiting_for_request = set()

    def worker(self):
        while self.running:
            with self.balancer.acquire_addr() as addr:
                self._fetch_and_run_one_job(addr)

    def _fetch_and_run_one_job(self, addr, cb_request_received=None):
        with socket(AF_INET, SOCK_STREAM) as s:
            self.sockets_waiting_for_request.add(s)
            try:
                s.connect(addr)

                s.setsockopt(SOL_SOCKET, SO_KEEPALIVE, 1)
                s.setsockopt(IPPROTO_TCP, TCP_KEEPIDLE, 1)
                s.setsockopt(IPPROTO_TCP, TCP_KEEPINTVL, 2)
                s.setsockopt(IPPROTO_TCP, TCP_KEEPCNT, 3)
            except Exception as e:
                LOGGER.error("Connect To Broker: %s", e)
                sleep(2)
                return

            try:
                Handler(s, addr, self, cb_request_received)
            except BaseException:
                LOGGER.exception("In WSGI request handling")
                sleep(1)
                return

    def get_app(self):
        return self.app

    def shutdown(self):
        LOGGER.info("shutting down workers...")
        self.running = False
        for s in list(self.sockets_waiting_for_request):
            s.shutdown(SHUT_RDWR)
            s.close()

    def poll_loop(self):
        """Uses app.wait_for_slot() spawn new workers dynamically. This is useful, for example to wait for
        enough CPU, memory, or disk space, to start next job.

        `wait_for_slot` is expected to block until it's time to poll for a new job. It is given a function
        `halt` which it should polle intermittently to check if this worker is undergoing `shutdown()`.

        It may return an object, if so `obj.request_received(environ)` will be called when a request is
        received and `obj.done()` when either the request is received, or if polling failed in which case
        `request_received()` was never called.

        See pothead.gating for some ready-made `wait_for_slot` implementations"""

        while self.running:
            callbacks = self.app.wait_for_slot(lambda: not self.running)
            if not self.running:
                break

            # This will block as long as we are waiting for requests from all broker-instances
            broker_instance_lease = self.broker_balancer.acquire_addr()

            Thread(
                target=self._run_one_poll, args=(broker_instance_lease, callbacks)
            ).start()

    def _run_one_poll(self, broker_instance_lease: ResourceLoan, callbacks):
        def request_received(environ):
            broker_instance_lease.release()
            getattr(callbacks, "request_received", lambda _: None)(environ)

        try:
            with broker_instance_lease as addr:
                self._fetch_and_run_one_job(addr, request_received)
        finally:
            if hasattr(callbacks, "done"):
                callbacks.done()


def install_term_handler(f):
    previous = []

    def on_term(_signal, _stack):
        f()
        for p in previous:
            p()

    p = signal(SIGTERM, on_term)
    if p:
        previous.append(p)


demo_app = simple_server.demo_app
demo_app.wait_for_slot = wait_for_idle_cpus(3)


if __name__ == "__main__":
    from argparse import ArgumentParser
    from importlib import import_module
    from logging import INFO, basicConfig
    from os import environ
    from sys import path

    path.insert(0, ".")

    def address(str):
        (host, port) = str.rsplit(":", 1)
        return (host, int(port))

    def func(str):
        (module, symbol) = str.rsplit(":", 1)
        module = import_module(module)
        return getattr(module, symbol)

    DEFAULT_WORKERS = int(environ.get("POTHEAD_WORKERS", 1))

    parser = ArgumentParser(description="Run WSGI app in sequential `worker` mode")
    parser.add_argument(
        "--connect",
        default="localhost:4040",
        type=address,
        help="Load Balancer Hub to connect to [host:port]",
    )
    job_control_group = parser.add_mutually_exclusive_group()
    job_control_group.add_argument(
        "--workers",
        default=DEFAULT_WORKERS,
        type=int,
        help="Number of worker Processes",
    )
    job_control_group.add_argument(
        "--poll-jobs",
        action="store_true",
        help="Use `app.wait_for_slot` to determine when to pull new jobs",
    )

    parser.add_argument(
        "app",
        nargs="?",
        default="pothead.worker:demo_app",
        type=func,
        help="The WSGI request handler to handle requests",
    )
    args = parser.parse_args()
    basicConfig(level=INFO)

    LOGGER.info("Initializing server for app %s", args.app)

    server = Server(args.connect, args.app)

    install_term_handler(server.shutdown)

    if args.poll_jobs:
        LOGGER.info(
            "Starting up job-polling dynamic workers connecting to %s", args.connect
        )
        server.poll_loop()
    else:
        LOGGER.info(
            "Starting up %d workers connecting to %s", args.workers, args.connect
        )
        workers = [Thread(target=server.worker) for _ in range(args.workers)]
        for worker in workers:
            worker.start()
        for worker in workers:
            worker.join()

    LOGGER.info("shut down")



from itertools import chain
from random import shuffle
from threading import Condition, Lock
from typing import Iterable
from math import inf


class Resource:
    def __init__(self, resource):
        self.resource = resource
        self.usage = 0

    def __cmp__(self, other):
        return self.usage.__cmp__(other.usage)

    def __lt__(self, other):
        return self.usage.__lt__(other.usage)


class ResourceLoan:
    def __init__(self, balancer, instance):
        self._lock = Lock()
        self._balancer = balancer
        self._instance = instance

    def __enter__(self):
        return self._instance.resource

    def release(self):
        with self._lock:
            if self._balancer:
                self._balancer._release(self._instance)
                self._balancer = None

    def __exit__(self, type, value, traceback):
        self.release()


class BoundedResourceBalancer:
    def __init__(self, bound=inf):
        self._active = list()
        self._ghosts = list()
        self._condition = Condition()
        self.bound = bound

    def _available(self):
        available = (r for r in self._active if r.usage < self.bound)
        return next(available, None)

    def active_count(self):
        return len(self._active)

    def acquire(self) -> ResourceLoan:
        with self._condition:
            winner = self._condition.wait_for(self._available)
            winner.usage += 1
            self._active.sort()
            return ResourceLoan(self, winner)

    def _release(self, instance):
        with self._condition:
            instance.usage -= 1
            self._active.sort()
            self._condition.notify_all()

    def provision(self, resources: Iterable[object]):
        with self._condition:
            new_resources = set(resources)

            active, ghosts = self._active, self._ghosts
            self._ghosts = []
            self._active = []

            for instance in chain(active, ghosts):
                try:
                    new_resources.remove(instance.resource)
                    self._active.append(instance)
                except KeyError:
                    if instance.usage > 0:
                        self._ghosts.append(instance)

            for new_resource in new_resources:
                self._active.append(Resource(new_resource))

            shuffle(self._active)
            self._active.sort()
            self._condition.notify_all()



#!/usr/bin/env python3

from asyncio import Queue, sleep
from asyncio import Future

from aiohttp import web
from aiohttp.client_reqrep import ClientRequest
from aiohttp.client_proto import ResponseHandler

jobs = Queue()


class MockConn:
    def __init__(self, protocol):
        self.protocol = protocol

    def release(self):
        pass


class ReverseHttp(ResponseHandler):
    def connection_made(self, transport):
        super().connection_made(transport)

        async def _deque():
            (req, future) = await jobs.get()
            req = ClientRequest(method=req.method, url=req.url, headers=req.headers)

            self.set_response_params()

            resp = await req.send(MockConn(self))
            await resp.start(MockConn(self))
            future.set_result(resp)

        self.task = self._loop.create_task(_deque())

    def connection_lost(self, exc):
        super().connection_lost(exc)
        self.task.cancel()


async def client_handler(request):
    response = Future()
    await jobs.put((request, response))
    resp = await response
    return web.Response(body=resp.content)


async def start_client_server(loop, port):
    runner = web.ServerRunner(web.Server(client_handler))
    await runner.setup()
    await web.TCPSite(runner, port=port).start()

    print("======= Serving clients on http://127.0.0.1:%d/ ======" % port)


async def start_worker_server(loop, port):
    await loop.create_server(lambda: ReverseHttp(loop), port=port)

    print("======= Serving workers on http://127.0.0.1:%d/ ======" % port)


async def main(loop, main_port, worker_port):
    await start_client_server(loop, main_port)
    await start_worker_server(loop, worker_port)

    print("======= Running ======")
    while True:
        await sleep(100 * 3600)


if __name__ == "__main__":
    from logging import basicConfig, DEBUG
    from asyncio import get_event_loop
    from argparse import ArgumentParser

    loop = get_event_loop()
    basicConfig(level=DEBUG)

    parser = ArgumentParser(description="Run PTTH Hub")
    parser.add_argument(
        "--main-port", default=8080, type=int, help="Port for public services"
    )
    parser.add_argument(
        "--worker-port",
        default=4040,
        type=int,
        help="Port where workers should connect",
    )
    args = parser.parse_args()

    try:
        loop.run_until_complete(main(loop, args.main_port, args.worker_port))
    except KeyboardInterrupt:
        pass
    loop.close()



import setuptools

with open("README.md", "r") as fh:
    long_description = fh.read()

setuptools.setup(
    name="POTHEAD",
    version="0.6.2",
    author="Ulrik Mikaelsson",
    author_email="ulrik.mikaelsson@gmail.com",
    description="A reverse-http proxy implementation for non-concurrent requests",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://gitlab.com/rawler/pothead",
    packages=setuptools.find_packages(),
    install_requires=["psutil", "werkzeug"],
    tests_require=["aiohttp"],
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: Apache Software License",
        "Operating System :: OS Independent",
    ],
)



from threading import Thread
from unittest import TestCase
from wsgiref import simple_server
from socket import socket, AF_INET, SOCK_STREAM, timeout as TimeoutError

from pothead.gating import wait_for_idle_cpus

from .worker import Server, LoadBalancer

demo_app = simple_server.demo_app
demo_app.wait_for_slot = wait_for_idle_cpus(0.1)


class WorkerConnection:
    def __init__(self, sock: socket):
        self.sock = sock

    def send_request(self):
        self.sock.sendall(b"GET / HTTP/1.1\r\n\r\n")


class DummyBroker:
    def __init__(self):
        self.sock = socket(AF_INET, SOCK_STREAM)
        self.sock.bind(("localhost", 0))
        self.sock.listen(1)

    def addr(self):
        return self.sock.getsockname()

    def accept(self, timeout) -> WorkerConnection:
        self.sock.settimeout(timeout)
        try:
            sock, _ = self.sock.accept()
            return WorkerConnection(sock)
        except TimeoutError:
            return None


def mock_balancer(addresses):
    class MockBalancer(LoadBalancer):
        def refresh(self):
            self._balancer.provision(addresses)

    return MockBalancer


class WorkerTest(TestCase):
    def test_load_balancing(self):
        broker_a, broker_b = (DummyBroker() for _ in range(2))

        worker = Server(
            ("", 0),
            demo_app,
            load_balancer=mock_balancer(map(DummyBroker.addr, (broker_a, broker_b))),
        )
        Thread(target=worker.poll_loop, daemon=True).start()

        connection_a = broker_a.accept(0.1)
        connection_b = broker_b.accept(0.1)

        assert broker_a.accept(0.1) is None
        assert broker_b.accept(0.1) is None

        connection_b.send_request()
        assert broker_b.accept(0.1) is not None

        connection_a.send_request()
        assert broker_a.accept(0.1) is not None
